@@ -1,19 +1,79 @@
 
-sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL, 
+sqldf <- function(x, stringsAsFactors = FALSE,
    row.names = FALSE, envir = parent.frame(), 
    method = getOption("sqldf.method"),
    file.format = list(), dbname, drv = getOption("sqldf.driver"), 
    user, password = "", host = "localhost",
-   dll = getOption("sqldf.dll"), connection = getOption("sqldf.connection")) {
+   dll = getOption("sqldf.dll"), connection = getOption("sqldf.connection"),
+   verbose = isTRUE(getOption("sqldf.verbose"))) {
 
+   as.POSIXct.numeric <- function(x, origin = "1970-01-01 00:00:00", ...)
+      base::as.POSIXct.numeric(x, origin = origin, ...)
    as.POSIXct.character <- function(x) structure(as.numeric(x),
 	class = c("POSIXt", "POSIXct"))
    as.Date.character <- function(x) structure(as.numeric(x), class = "Date")
+   as.Date2 <- function(x) UseMethod("as.Date2")
+   as.Date2.character <- function(x) base:::as.Date.character(x)
    as.Date.numeric <- function(x, origin = "1970-01-01", ...) base::as.Date.numeric(x, origin = origin, ...)
    as.dates.character <- function(x) structure(as.numeric(x), class = c("dates", "times"))
    as.times.character <- function(x) structure(as.numeric(x), class = "times")
 
 
+   # nam2 code is duplicated above.  Needs to be factored out.
+   backquote.maybe <- function(nam) {
+		if (drv == "h2") { nam
+		} else if (drv == "mysql") { nam
+		} else if (drv == "pgsql") { nam
+		} else {
+			if (regexpr(".", nam, fixed = TRUE)) {
+				paste("`", nam, "`", sep = "")
+			} else nam
+		}
+	}
+
+   name__class <- function(data, ...) {
+	if (is.null(data)) return(data)
+	cls <- sub(".*__([^_]+)|.*", "\\1", names(data))
+	f <- function(i) {
+		if (cls[i] == "") { 
+			data[[i]] 
+		} else {
+			fun_name <- paste("as", cls[i], sep = ".")
+			fun <- mget(fun_name, env = environment(), 
+				mode = "function", ifnotfound = NA, inherit = TRUE)[[1]]
+			if (identical(fun, NA)) data[[i]] else {
+				# strip _class off name
+				names(data)[i] <<- sub("__[^_]+$", "", names(data)[i])
+				fun(data[[i]])
+			}
+		}
+	}
+	data[] <- lapply(1:NCOL(data), f)
+	data
+   }
+
+	colClass <- function(data, cls) {
+	  if (is.null(data)) return(data)
+	  if (is.list(cls)) cls <- unlist(cls)
+	  cls <- rep(cls, length = length(data))
+	  f <- function(i) {
+		if (cls[i] == "") { 
+			data[[i]] 
+		} else {
+			fun_name <- paste("as", cls[i], sep = ".")
+			fun <- mget(fun_name, env = environment(), 
+				mode = "function", ifnotfound = NA, inherit = TRUE)[[1]]
+			if (identical(fun, NA)) data[[i]] else {
+				# strip _class off name
+				names(data)[i] <<- sub("__[^_]+$", "", names(data)[i])
+				fun(data[[i]])
+			}
+		 }
+	   }
+	data[] <- lapply(1:NCOL(data), f)
+	data
+   }
+
 	overwrite <- FALSE
 
 	request.open <- missing(x) && is.null(connection)
@@ -23,6 +83,11 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 
 	dfnames <- fileobjs <- character(0)
 
+
+	if (!is.list(method)) method <- list(method, NULL)
+	to.df <- method[[1]]
+	to.db <- method[[2]]
+
 	# if exactly one of x and connection are missing then close on exit
 	if (request.close || request.nocon) { 
 
@@ -30,15 +95,37 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 			dbPreExists <- attr(connection, "dbPreExists")
 			dbname <- attr(connection, "dbname")
     		if (!missing(dbname) && !is.null(dbname) && dbname == ":memory:") {
+				if (verbose) {
+					cat("sqldf: dbDisconnect(connection)\n")
+				}
 				dbDisconnect(connection)
     		} else if (!dbPreExists && drv == "sqlite") {
     			# data base not pre-existing
+				if (verbose) {
+					cat("sqldf: dbDisconnect(connection)\n")
+					cat("sqldf: file.remove(dbname)\n")
+				}
     			dbDisconnect(connection)
     			file.remove(dbname)
     		} else {
     			# data base pre-existing
-    			for (nam in dfnames) dbRemoveTable(connection, nam)
-    			for (fo in fileobjs) dbRemoveTable(connection, fo)
+
+    			for (nam in dfnames) {
+					nam2 <- backquote.maybe(nam)
+					if (verbose) {
+						cat("sqldf: dbRemoveTable(connection, ", nam2, ")\n")
+					}
+					dbRemoveTable(connection, nam2)
+				}
+    			for (fo in fileobjs) {
+					if (verbose) {
+						cat("sqldf: dbRemoveTable(connection, ", fo, ")\n")
+					}
+					dbRemoveTable(connection, fo)
+				}
+				if (verbose) {
+					cat("sqldf: dbDisconnect(connection)\n")
+				}
     			dbDisconnect(connection)
     		}
     	})
@@ -61,12 +148,18 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
     
 		drv <- tolower(drv)
     	if (drv == "mysql") {
+			if (verbose) cat("sqldf: m <- dbDriver(\"MySQL\")\n")
     		m <- dbDriver("MySQL")
+			if (missing(dbname) || is.null(dbname)) {
+				dbname <- getOption("RMySQL.dbname")
+				if (is.null(dbname)) dbname <- "test"
+			}
     		connection <- if (missing(dbname) || dbname == ":memory:") { 
     				dbConnect(m) 
     			} else dbConnect(m, dbname = dbname)
     			dbPreExists <- TRUE
 		} else if (drv == "pgsql") {
+			if (verbose) cat("sqldf: m <- dbDriver(\"pgSQL\")\n")
     		m <- dbDriver("pgSQL")
 			if (missing(dbname) || is.null(dbname)) {
 				dbname <- getOption("RpgSQL.dbname")
@@ -78,7 +171,9 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 			# jar.file <- "C:\\Program Files\\H2\\bin\\h2.jar"
 			# jar.file <- system.file("h2.jar", package = "H2")
 			# m <- JDBC("org.h2.Driver", jar.file, identifier.quote = '"')
-			m <- H2()
+			if (verbose) cat("sqldf: m <- dbDriver(\"H2\")\n")
+			# m <- H2()
+			m <- dbDriver("H2")
     		if (missing(dbname) || is.null(dbname)) dbname <- ":memory:"
     		dbPreExists <- dbname != ":memory:" && file.exists(dbname)
 			connection <- if (missing(dbname) || dbname == ":memory:") {
@@ -89,15 +184,18 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 					dbConnect(m, jdbc.string)
 				}
 		} else {
+			if (verbose) cat("sqldf: m <- dbDriver(\"SQLite\")\n")
     		m <- dbDriver("SQLite")
     		if (missing(dbname)) dbname <- ":memory:"
     		dbPreExists <- dbname != ":memory:" && file.exists(dbname)
 
 			# search for spatialite extension on PATH and, if found, load it
-			if (is.null(getOption("sqldf.dll"))) {
-				dll <- Sys.which("libspatialite-1.dll")
-				if (dll != "") options(sqldf.dll = dll) else options(sqldf.dll = FALSE)
-			}
+			# if (is.null(getOption("sqldf.dll"))) {
+			#	dll <- Sys.which("libspatialite-2.dll")
+			#	if (dll == "") dll <- Sys.which("libspatialite-1.dll")
+			#	dll <- if (dll == "") FALSE else normalizePath(dll)
+			#	options(sqldf.dll = dll)
+			# }
 			dll <- getOption("sqldf.dll")
 			if (length(dll) != 1 || identical(dll, FALSE) || nchar(dll) == 0) {
 				dll <- FALSE
@@ -107,13 +205,22 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 			options(sqldf.dll = dll)
 
 			if (!identical(dll, FALSE)) {
+				if (verbose) {
+					cat("sqldf: connection <- dbConnect(m, dbname = \"", dbname, 
+						"\", loadable.extensions = TRUE\n", sep = "")
+					cat("sqldf: select load_extension('", dll, "')\n", sep = "")
+				}
 				connection <- dbConnect(m, dbname = dbname, 
 					loadable.extensions = TRUE)
 				s <- sprintf("select load_extension('%s')", dll)
 				dbGetQuery(connection, s)
-			} else connection <- dbConnect(m, dbname = dbname)
+			} else {
+				cat("sqldf: connection <- dbConnect(m, dbname = \"", dbname, "\")\n", sep = "")
+				connection <- dbConnect(m, dbname = dbname)
+			}
 			# if (require("RSQLite.extfuns")) init_extensions(connection)
 			# load extension functions from RSQLite.extfuns
+			if (verbose) cat("sqldf: init_extensions(connection)\n")
 			init_extensions(connection)
     	}
 		attr(connection, "dbPreExists") <- dbPreExists
@@ -156,15 +263,12 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 		}
 		# check if the nam2 processing works with MySQL
 		# if not then ensure its only applied to SQLite
-		nam2 <- if (drv == "h2") { nam
-		} else if (drv == "pgsql") { nam
-		} else {
-			if (regexpr(".", nam, fixed = TRUE)) {
-				paste("`", nam, "`", sep = "")
-			} else nam
-		}
-		dbWriteTable(connection, nam2, as.data.frame(get(nam, envir)), 
-			row.names = row.names)
+		DF <- as.data.frame(get(nam, envir))
+		if (!is.null(to.db) && is.function(to.db)) DF <- to.db(DF)
+		nam2 <- backquote.maybe(nam)
+		# if (verbose) cat("sqldf: writing", nam2, "to database\n")
+		if (verbose) cat("sqldf: dbWriteTable(connection, '", nam2, "', ", nam, ", row.names = ", row.names, ")\n", sep = "")
+		dbWriteTable(connection, nam2, DF, row.names = row.names)
 	}
 
 	# process file objects
@@ -190,7 +294,22 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 			args$filter <- NULL
 			Filename.tmp <- tempfile()
 			args$value <- Filename.tmp
-			cmd <- sprintf("%s < %s > %s", filter, Filename, Filename.tmp)
+			# for filter = list(cmd, x = ...) replace x in cmd with
+			# temporary filename for a file created to hold ...
+			filter.subs <- filter[-1]
+			# filter subs contains named elements of filter
+			filter.subs <- filter.subs[sapply(names(filter.subs), nzchar)]
+			filter.nms <- names(filter.subs)
+			# create temporary file names
+			filter.tempfiles <- sapply(filter.nms, tempfile)
+			cmd <- filter[[1]]
+			# write out temporary file & substitute temporary file name into cmd
+			for(nm in filter.nms) {
+				cat(filter.subs[[nm]], file = filter.tempfiles[[nm]])
+				cmd <- gsub(nm, filter.tempfiles[[nm]], cmd, fixed = TRUE)
+			}
+			cmd <- sprintf('%s < "%s" > "%s"', cmd, Filename, Filename.tmp)
+
 			# on Windows platform preface command with cmd /c 
 			if (.Platform$OS == "windows") {
 				cmd <- paste("cmd /c", cmd)
@@ -211,32 +330,51 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 				}
 				}
 			}
+			if (verbose) cat("sqldf: system(\"", cmd, "\")\n  ", sep = "")
 			system(cmd)
+			for(fn in filter.tempfiles) file.remove(fn)
 		}
+		if (verbose) cat("sqldf: dbWriteTable(", args, ")\n")
 		do.call("dbWriteTable", args)
 	}
 
 	# SQLite can process all statements using dbGetQuery.  
 	# Other databases process select/call/show with dbGetQuery and other 
 	# statements with dbSendQuery.
-	if (drv == "sqlite") {
-		for(xi in x) rs <- dbGetQuery(connection, xi)
+	if (drv == "sqlite" || drv == "mysql") {
+		for(xi in x) {
+			if (verbose) {
+				cat("sqldf: dbGetQuery(connection, '", xi, "')\n", sep = "")
+			}
+			rs <- dbGetQuery(connection, xi)
+		}
 	} else {
 		for(i in seq_along(x)) {
 			if (length(words.[[i]]) > 0) {
 				dbGetQueryWords <- c("select", "show", "call", "explain")
 				if (tolower(words.[[i]][1]) %in% dbGetQueryWords) {
+					if (verbose) {
+						cat("sqldf: dbGetQuery(connection, '", x[i], "')\n", sep = "")
+					}
 					rs <- dbGetQuery(connection, x[i])
 				} else {
+					if (verbose) {
+						cat("sqldf: dbSendUpdate:", x[i], "\n")
+					}
 					rs <- dbSendUpdate(connection, x[i])
 				}
 			}
 		}
 	}
 
-	# get result back
-	if (is.null(method)) method <- "auto"
-	if (match.arg(method, c("auto", "raw")) == "raw") return(rs)
+	if (is.null(to.df)) to.df <- "auto"
+    if (is.function(to.df)) return(to.df(rs))
+	# to.df <- match.arg(to.df, c("auto", "raw", "name__class"))
+	if (identical(to.df, "raw")) return(rs)
+	if (identical(to.df, "name__class")) return(do.call("name__class", list(rs)))
+	if (!identical(to.df, "nofactor") && !identical(to.df, "auto")) {
+		return(do.call("colClass", list(rs, to.df)))
+	}
 	# process row_names
 	rs <- if ("row_names" %in% names(rs)) {
 		if (identical(row.names, FALSE)) {
@@ -257,44 +395,86 @@ sqldf <- function(x, stringsAsFactors = TRUE, col.classes = NULL,
 	#
 	# For each column name in the result, match it against each column name
 	# of each data frame in envir.
-	# 
 	#
-	tab <- do.call("rbind", lapply(dfnames, function(dfname) {
-		df <- get(dfname, envir)
-		cbind(dfname, colnames(df))
-	}))
-	# column names which are duplicated
-	dup <- tab[,2][duplicated(tab[,2])]
+	# tab has one row for each column in each data frame with columns being: 
+	# (1) data frame name, 
+	# (2) column name, 
+	# (3) class vector concatenated using toString,
+	# (4) levels concatenated using toString
+	#
+	tab <- do.call("rbind", 
+		lapply(dfnames, 
+			# calculate tab for one data frame
+			function(dfname) {
+				df <- get(dfname, envir)
+				nms <- names(df)
+				do.call("rbind", 
+					lapply(seq_along(df), 
+						# calculate row in tab for one column of one data frame
+						function(j) {
+							column <- df[[j]]
+							cbind(dfname, nms[j], toString(class(column)), 
+								toString(levels(column)))
+						}
+					)
+				)
+			}
+		)
+	)
+			
+	# each row is a unique column name/class vector/levels combo
+	tabu <- unique(tab[,-1,drop=FALSE])
 
-	f <- function(i) {
+	# dup is vector of column names that appear more than once in tabu.
+	# Such column names have conflicting classes or levels and therefore
+	# cannot form the basis of unique class and level assignments.
+	dup <- unname(tabu[duplicated(tabu[,1]), 1])
+
+	# cat("tab:\n")
+	# print(tab)
+	# cat("tabu:\n")
+	# print(tabu)
+	# cat("dup:\n")
+	# print(dup)
+
+	auto <- function(i) {
 		cn <- colnames(rs)[[i]]
 		if (! cn %in% dup && 
 			(ix <- match(cn, tab[, 2], nomatch = 0)) > 0) {
 			df <- get(tab[ix, 1], envir)
-			if (inherits(df[[cn]], "ordered"))
-				return(as.ordered(factor(rs[[cn]], 
-					levels = levels(df[[cn]]))))
-			else if (inherits(df[[cn]], "factor"))
-				return(factor(rs[[cn]], 
-					levels = levels(df[[cn]])))
-			else if (inherits(df[[cn]], "POSIXct"))
-				return(as.POSIXct(rs[[cn]]))
-			else if (identical(class(df[[cn]]), "times")) 
+			if (inherits(df[[cn]], "ordered")) {
+				if (identical(to.df, "auto")) {
+					u <- unique(rs[[i]])
+					levs <- levels(df[[cn]])
+					if (all(u %in% levs))
+						return(factor(rs[[i]], levels = levels(df[[cn]]), 
+							order = TRUE))
+					else return(rs[[i]])
+				} else return(rs[[i]])
+			} else if (inherits(df[[cn]], "factor")) {
+				if (identical(to.df, "auto")) {
+					u <- unique(rs[[i]])
+					levs <- levels(df[[cn]])
+					if (all(u %in% levs))
+						return(factor(rs[[i]], levels = levels(df[[cn]])))
+					else return(rs[[i]])
+				} else return(rs[[i]])
+			} else if (inherits(df[[cn]], "POSIXct"))
+				return(as.POSIXct(rs[[i]]))
+			else if (inherits(df[[cn]], "times")) 
 				return(times(df[[cn]]))
 			else {
 				asfn <- paste("as", 
 					class(df[[cn]]), sep = ".")
 				asfn <- match.fun(asfn)
-				return(asfn(rs[[cn]]))
+				return(asfn(rs[[i]]))
 			}
 		}
-		if (stringsAsFactors) 
-			if (is.character(rs[[i]]))
-				factor(rs[[i]])
-			else rs[[i]]
+		if (stringsAsFactors && is.character(rs[[i]])) factor(rs[[i]])
+		else rs[[i]]
 	}
 	# debug(f)
-	rs2 <- lapply(seq_along(rs), f)
+	rs2 <- lapply(seq_along(rs), auto)
 	rs[] <- rs2
 	rs
 }
@@ -329,7 +509,7 @@ read.csv2.sql <- function(file, sql = "select * from file",
 		else "tr , ."
 	}
 
-	read.csv.sql(file = file, sql = sql, header = header, sep = sep, 
+read.csv.sql(file = file, sql = sql, header = header, sep = sep, 
 		row.names = row.names, eol = eol, skip = skip, filter = filter, 
 		dbname = dbname, drv = drv)
 }
